# -*- coding: utf-8 -*-
"""Registry of available tasks."""
import inspect
from importlib import import_module
from typing import Any
from celery._state import get_current_app
from celery.exceptions import NotRegistered, InvalidTaskError
from celery.types import TaskT

__all__ = ['TaskRegistry']


class TaskRegistry(dict):
    """Map of registered tasks."""

    NotRegistered = NotRegistered

    def __missing__(self, key: str) -> Any:
        raise self.NotRegistered(key)

    def register(self, task: TaskT) -> None:
        """Register a task in the task registry.

        The task will be automatically instantiated if not already an
        instance. Name must be configured prior to registration.
        """
        if task.name is None:
            raise InvalidTaskError(
                'Task class {0!r} must specify .name attribute'.format(
                    type(task).__name__))
        self[task.name] = inspect.isclass(task) and task() or task

    def unregister(self, name: str) -> None:
        """Unregister task by name.

        Arguments:
            name (str): name of the task to unregister, or a
                :class:`celery.task.base.Task` with a valid `name` attribute.

        Raises:
            celery.exceptions.NotRegistered: if the task is not registered.
        """
        try:
            self.pop(getattr(name, 'name', name))
        except KeyError:
            raise self.NotRegistered(name)


def _unpickle_task(name: str) -> TaskT:
    return get_current_app().tasks[name]


def _unpickle_task_v2(name: str, module: str = None) -> TaskT:
    if module:
        import_module(module)
    return get_current_app().tasks[name]
